"""SQL handler for reading from SQL service.

>>> df = read_sql("rich_REDSHIFT", "SELECT 1;")
"""
from os import environ as env

from jinja2 import Environment
import pandas as pd
import requests


def _get_endpoint_payload():
    """Create route and key headers for API."""
    endpoint = env['SQL_URL']
    key = env['SQL_KEY']
    headers = {'x-api-key': key}
    return dict(headers=headers, route=endpoint)


def read_sql(backend: str, sql: str, sql_params=None) -> str:
    """Given a backend, sql and params, run the sql."""
    payload = _get_endpoint_payload()
    route = payload['route'] + f'/view/{backend}'

    sql_params = sql_params if sql_params else {}
    rendered_sql = Environment().from_string(sql).render(**sql_params)

    params = dict(sql=rendered_sql)
    response = requests.get(
        route,
        headers=payload['headers'],
        params=params,
    )
    response.raise_for_status()
    return pd.read_json(response.text)


def list_backends():
    """Retreive all backends registered."""
    payload = _get_endpoint_payload()
    route = payload['route'] + f'/backends'

    response = requests.get(route, headers=payload['headers'])
    response.raise_for_status()
    return response.json()



#! /usr/bin/env python
"""Package for interacting with pandas extension modules.

Any additional repos that may require client-side libs to do
data manipulation.
"""
from setuptools import find_packages, setup

import codecs
import os
import re


###############################################################
NAME = "pandas_ext"
PACKAGES = find_packages(exclude=('tests', 'docs'))
META_PATH = os.path.join("pandas_ext", "__init__.py")
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 2.7',
]
##############################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


README = read('README.md')
META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """

    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError(f"Unable to find __{meta}__ string.")


setup(
    name=NAME,
    description=find_meta("description"),
    author=find_meta("author"),
    author_email=find_meta("email"),
    url=find_meta("uri"),
    version=find_meta("version"),
    license=find_meta("license"),
    install_requires=read("requirements/requirements.in"),
    long_description=README,
    packages=PACKAGES,
    classifiers=CLASSIFIERS,
)



"""
pandas dtype hierarchy.

[numpy.generic,
[[numpy.number,
[[numpy.integer,
[[numpy.signedinteger,
[numpy.int8,
numpy.int16,
numpy.int32,
numpy.int64,
numpy.int64,
numpy.timedelta64]],
[numpy.unsignedinteger,
[numpy.uint8,
numpy.uint16,
numpy.uint32,
numpy.uint64,
numpy.uint64]]]],
[numpy.inexact,
[[numpy.floating,
[numpy.float16, numpy.float32, numpy.float64, numpy.float128]],
[numpy.complexfloating,
[numpy.complex64, numpy.complex128, numpy.complex256]]]]]],
[numpy.flexible,
[[numpy.character, [numpy.bytes_, numpy.str_]],
[numpy.void, [numpy.record]]]],
numpy.bool_,
numpy.datetime64,
numpy.object_]]

redshift spectrum data types:
Data Type   Aliases     Description
SMALLINT    INT2    Signed two-byte integer
INTEGER     INT, INT4   Signed four-byte integer
BIGINT  INT8    Signed eight-byte integer
DECIMAL     NUMERIC     Exact numeric of selectable precision
REAL    FLOAT4  Single precision floating-point number
DOUBLE PRECISION    FLOAT8, FLOAT   Double precision floating-point number
BOOLEAN     BOOL    Logical Boolean (true/false)
CHAR    CHARACTER, NCHAR, BPCHAR    Fixed-length character string
VARCHAR     CHARACTER VARYING, NVARCHAR, TEXT   Variable-length character string with a user-defined limit
DATE        Calendar date (year, month, day)
TIMESTAMP   TIMESTAMP WITHOUT TIME ZONE     Date and time (without time zone)
TIMESTAMPTZ     TIMESTAMP WITH TIME ZONE    Date and time (with time zone) 
"""
import pandas as pd
from pandas.api.types import pandas_dtype
import numpy as np


def dtype_to_spectrum(dtype):
    """convert pandas dtype to equivalent redshift spectrum schema column value."""
    try:
        return {
            pandas_dtype(np.float64): 'FLOAT8',
            pandas_dtype(np.object): 'VARCHAR(8192)',
            pandas_dtype(np.int64): 'INT8',
            pandas_dtype(np.bool): 'BOOL',
            pandas_dtype(np.datetime64): 'TIMESTAMP',
            pandas_dtype('<M8[s]'): 'TIMESTAMP'
        }[dtype]
    except KeyError:
        return 'TEXT'


def schema_from_df(df: pd.DataFrame):
    dtype_map = df.dtypes.to_dict()
    return ',\n'.join(
            [f'"{col}" {dtype_to_spectrum(dtype)}'
                for col, dtype in dtype_map.items()
            ]
    )



from inspect import cleandoc

import pandas as pd

from pandas_ext.common.utils import today
from pandas_ext.parquet import to_parquet
from pandas_ext.sqla_utils import schema_from_df


def _get_file_format_serde(file_format: str) -> str:
    return dict(
        parquet='org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe'
    )[file_format]


def _build_s3_stream_path(
    bucket,
    stream,
    file_format,
    partition,
    partition_value
):
    return (f's3://{bucket}/{stream}/ext={file_format}/'
            f'{partition}={partition_value}/{stream}.snappy'
            ).lower()


def _create_schema_alias_statement(
    schema_alias: str,
    schema: str,
    table: str
) -> str:
    return cleandoc(f"""
        CREATE VIEW "{schema_alias}"."{table}" AS
        SELECT * FROM "{schema}"."{table}"
        WITH NO SCHEMA BINDING;
    """
                    )


def _create_partition_statement(
    schema: str,
    bucket: str,
    stream: str,
    file_format: str='parquet',
    partition: str='dt',
    partition_value: str=''
) -> str:
    partition_value = (
        partition_value if partition_value else
        today()
    )
    s3_path = f's3://{bucket}/{stream}/ext={file_format}/{partition}={partition_value}/'.lower()
    return cleandoc(f"""
        ALTER TABLE "{schema}"."{stream}_{file_format}"
        ADD PARTITION ({partition}='{partition_value}')
        LOCATION '{s3_path}'
        ;"""
                    )


def _external_table_exists_statement(
    schema: str,
    table: str
) -> str:

    return cleandoc(f"""
        SELECT distinct(schemaname || tablename) as schema_table
        FROM SVV_EXTERNAL_COLUMNS
        WHERE schemaname || tablename = '{schema}{table}'
        ;"""
                    )


def _create_external_table_statement(
    schema: str,
    table: str,
    columns: str,
    bucket: str,
    stream: str,
    file_format: str='parquet',
    partition: str='dt',
    partition_type: str='date',
    partition_value: str=''
) -> str:
    serde = _get_file_format_serde(file_format)
    upper_file_format = file_format.upper()
    s3_path = f's3://{bucket}/{stream}/ext={file_format}/'.lower()

    return cleandoc(f"""
        CREATE EXTERNAL TABLE "{schema}"."{table}_{file_format}" (
        {columns}
        )PARTITIONED BY ({partition} {partition_type})
        ROW FORMAT SERDE '{serde}'
        STORED AS {upper_file_format}
        LOCATION '{s3_path}'
        ;"""
                    )


def to_spectrum(
    df: pd.DataFrame,
    table: str,
    schema: str,
    bucket: str,
    schema_alias: str='',
    stream: str='',
    file_format: str='parquet',
    partition: str='dt',
    partition_type: str='date',
    partition_value: str='',
    conn: str='',
    **kwargs
) -> str:
    """Sends your dataframe to Spectrum for use in Athena/Redshift/Looker/etc

       Currently we only print out the statements as only the owner of the
       external schema can actually run the CREATE EXTERNAL TABLE statement.

       df: pandas Dataframe
       schema: external table schema
       table: table name as it appears in Spectrum
       bucket: s3 bucket
       schema_alias: If you want to create an alternate path to your schema
       stream: Defaults to table if not provided.
       file_format: Defaults to parquet and may expand to avro.
       partition: Defaults to dt, which is short for the date.
       partition_type: The data type declaration of the partition value.
       partition_value: Defaults to todays date.
       conn: A valid sqlalchemy string to connect to spectrum.
       kwargs: kwargs you want to pass to `to_parquet()` call
    """

    columns = schema_from_df(df)
    stream = stream if stream else table
    external_table_statement = _create_external_table_statement(
        schema=schema,
        table=table,
        columns=columns,
        bucket=bucket,
        stream=stream,
        file_format=file_format,
        partition=partition,
        partition_type=partition_type
    )
    alias_statement = (
        '' if not schema_alias else
        _create_alias_statement(schema_alias, schema, table)
    )
    partition_value = (
        partition_value if partition_value else
        today()
    )
    partition_statement = _create_partition_statement(
        schema=schema,
        bucket=bucket,
        stream=stream,
        partition=partition,
        partition_value=partition_value
    )
    create_statement = (''.join([
        external_table_statement,
        alias_statement,
        partition_statement
    ]))
    print(create_statement)

    s3_path = _build_s3_stream_path(
        bucket,
        stream,
        file_format,
        partition,
        partition_value)
    print(f'SELECT COUNT(*) FROM "{schema}"."{table}_{file_format}";')
    print(f"df_{table} = read_parquet('{s3_path}')")

    if conn:
        print(s3_path)
        to_parquet(df, s3_path, **kwargs)
        from sqlalchemy import create_engine
        engine = create_engine(conn, execution_options=dict(autocommit=True))

        schema_table_statement = _external_table_exists_statement(
            schema, table)
        table_exists_data = pd.read_sql_query(schema_table_statement, conn)
        if not len(table_exists_data):
            # table doesn't exist so create it.
            engine.execute(create_statement)
        engine.execute(partition_statement)
